/* ZiNiCalculator */
/* version 0.1 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define ZNC_VERSION 0.3

struct cell
{
	/* cell value (read* makes it after initializing by clearboard) */
	int mine;
	/* cell properties (initboard makes these things) */
	int opening; /* 0 iff cell is non-opening 3bv */
	int opening2;	/* one cell may belong to 2 openings */
	int number;
	/* cell position (initboard inits these too) */
	int rb,re,cb,ce; /* these things are there to determine cell neighbours */
	/* cell current state (clearboard&restartboard init it) */
	int opened;
	int flagged;
	/* ZiNi things */
	int premium;
};

typedef struct cell cell;

cell* board; /* board[column*height+row] */
int width;
int height;
int size;	/* =width*height */
int mines;

int openings;
int bbbv;
int closed_cells;

int zini;

/* board file */
FILE* f;

/* general settings */
/* they are here to avoid their passing to functions, subfunctions, etc */
int human,rz,hitops,nobrd,a3,turns,infos,priter,s_iter;
int n_iter;

/* storage for maximal premium cells for random zini */
int* maxpr;


void printhelp()
{
	printf("%s\n","Usage: zinicalc [options [number_of_iterations]] <input_file>");
	printf("%s%.1f\n"," ZiNi calculator v",ZNC_VERSION);
	printf("%s"," ZiNi is an heuristic algorithm proposed by Elmar Zimmermann and Christoph\nNikolaus Marx");
	printf("%s"," to estimate minimum number of clicks needed to finish the board.\n");
	printf("%s\n","For details see http://www.minesweeper.info/forum/viewtopic.php?t=70");
	printf("%s\n"," Available options:");
	printf("%s\n","  -h calculate Human ZiNi");
	printf("%s\n","  -r calculate Random ZiNi");
	printf("%s\n","  -o hit openings before ZiNi applying");
	printf("%s\n","  -t pause between iterations");
	printf("%s\n","  -s suppress clicks output");
	printf("%s\n","  -b suppress board output");
	printf("%s\n","  -a print adjacent 3bv for cells");
	printf("%s\n","  -i suppress iterations output");
	printf("%s\n"," Supported formats:");
	printf("%s\n","  - MBF (ABF)");
	printf("%s\n","  - Clone MVF");
	printf("%s\n","  - Arbiter AVF");
	printf("%s\n","  - Vienna RMV and UMF");
}

void pause()
{
	while(getchar()!='\n');
}

void error(const char* s)
{
	printf("%s\n",s);
	printf("Press enter to exit\n");
	pause();
	exit(1);
}

void clearboard()
{
	int i;
	closed_cells=size;
	for(i=0;i<size;++i)
		board[i].mine=
		board[i].opened=board[i].flagged=board[i].opening=board[i].opening2=0;
}

void restartboard()
{
	int i;
	closed_cells=size;
	for(i=0;i<size;++i)
		board[i].opened=board[i].flagged=board[i].opening=board[i].opening2=0;
}

void readmines(int add) /* column then row */
{
	unsigned char c;
	int pos;
	int i;
	for(i=0;i<mines;++i)
	{
		c=fgetc(f);pos=(c+add)*height;
		c=fgetc(f);pos+=c+add;
		if(pos>=size || pos<0) error("Invalid mine position");
		board[pos].mine=1;
	}
}
void readminesr(int add)	/* rilian mines are so rilian */
{				/* row then column */
	unsigned char c;
	int pos;
	int i;
	for(i=0;i<mines;++i)
	{
		c=fgetc(f);pos=c+add;
		c=fgetc(f);pos+=(c+add)*height;
		board[pos].mine=1;
	}
}
void readboard(int add)
{
	/* since the board is pretty the same for mbf, mvf and others... */
	unsigned char c;
	c=fgetc(f);width=c;
	c=fgetc(f);height=c;
	board=(cell*)malloc((size=width*height)*sizeof(cell));
	clearboard();
	c=fgetc(f);mines=c<<8;
	c=fgetc(f);mines+=c;
	readmines(add);
}
void readmbf()
{
	printf("File seems to be MBF...\n");
	readboard(0);
}
void readoldmvf()
{
	printf("File seems to be pre-0.97 MVF...\n");
	fseek(f,0,SEEK_SET);
	readboard(0);
}
void readmvf()
{
	/* damn Rodrigo's fickleness... This section is going to be BIG! */
	unsigned char c;
	c=fgetc(f);
	if(c!=17)	/* it's old MVF without header */
		readoldmvf();
	else	/* look at key byte */
	{
		c=fgetc(f);
		if(c!='M') {printf("hehe\n");readoldmvf();return;} /* i suppose height<'M'=77 */
		fseek(f,27,SEEK_SET);
		c=fgetc(f);
		switch(c)
		{
		case '5':
			printf("File seems to be 0.97 MVF...\n");
			fseek(f,97,SEEK_SET);	/* 97 for 0.97... lol */
			readboard(-1);
			break;
		case '6':
			printf("File seems to be 2006 MVF...\n");
			fseek(f,84,SEEK_SET);
			readboard(-1);
			break;
		case '7':
			printf("File seems to be 2007 MVF...\n");
			fseek(f,84,SEEK_SET);
			readboard(-1);
			break;
		default:
			error("I don't know such MVF!");
		}
	}
}
void readavf()
{
	/* damn rilian's strangeness... This section is going to be odd! */
	unsigned char c;
	c=fgetc(f);
	printf("File seems to be 0.%d AVF\n",(int)c);
	fseek(f,5,SEEK_SET);
	c=fgetc(f);
	switch(c)
	{
	case 3:
		width=height=8;mines=10;break;
	case 4:
		width=height=16;mines=40;break;
	case 5:
		width=30;height=16;mines=99;break;
	case 6:
		c=fgetc(f);width=c+1;
		c=fgetc(f);height=c+1;
		c=fgetc(f);mines=c<<8;
		c=fgetc(f);mines+=c;
	}
	board=(cell*)malloc((size=width*height)*sizeof(cell));
	clearboard();
	readminesr(-1);
}
void seekchar(unsigned char cc)
{
	unsigned char c=0;
	while(c!=cc) c=fgetc(f);
}
void readrmv()
{
	/* damn Christoph's inventiveness... And where's the board??? */
	unsigned char c;
	unsigned char l;
	int i;
	c=fgetc(f);c=fgetc(f);
	if(c=='u')
		printf("File seems to be UMF\n");
	else if(c=='r')
		printf("File seems to be RMV\n");
	/* let's find string "Marx"0x00 0x01 */
	while(1)
	{
		seekchar('M');
		c=fgetc(f);if(c!='a') continue;
		c=fgetc(f);if(c!='r') continue;
		c=fgetc(f);if(c!='x') continue;
		c=fgetc(f);if(c!=0) continue;
		c=fgetc(f);if(c!=1) continue;
		break;
	}
	/* then rmv and umf contain nick, we'll go through it */
	l=fgetc(f);for(i=0;i<l;++i) c=fgetc(f);
	/* then go 4 pretty random bytes */
	for(i=0;i<4;++i) c=fgetc(f);
	/* FINALLY!!! it should be the board */
	readboard(0);
}

void open(int index)
{
	int rr,cc;
	board[index].opened=1;
	++board[index].premium;	/* it's now opened! */
	if(!board[index].opening) /* if it's non-opening 3bv */
		for(rr=board[index].rb;rr<=board[index].re;++rr)
			for(cc=board[index].cb;cc<=board[index].ce;++cc)
				--board[cc*height+rr].premium; /* reduce premium of neighbours */
	--closed_cells;
}
void reveal(int index)
{
	if(board[index].opened) return;	/* wtf?! it was opened already!! */
	if(board[index].flagged) return;

	if(board[index].number)
		open(index);
	else /* damn it's an opening!!! */
	{
		int op=board[index].opening;
		int i;
		for(i=0;i<size;++i)
		{
			if(board[i].opening2==op ||
				board[i].opening==op)
			{
				if(!board[i].opened) open(i);
				--board[i].premium;	/* reduce adjacent 3bv, nearby opening opened */
			}
		}
	}
}
void info(const char* inf,int index)
{
	if(infos)
		printf("%s : row %d column %d\n",inf,index%height+1,index/height+1);
}
void click(int index)
{
	info("Left click",index);
	reveal(index);
	++zini;
}
void flag(int index)
{
	int rr,cc;
	if(board[index].flagged) return;
	info("Flag",index); 
	++zini;
	board[index].flagged=1;
	for(rr=board[index].rb;rr<=board[index].re;++rr)
		for(cc=board[index].cb;cc<=board[index].ce;++cc)
			++board[cc*height+rr].premium; /* increase premium of neighbours */
}
void flagaround(int index)
{
	int rr,cc;
	for(rr=board[index].rb;rr<=board[index].re;++rr)
		for(cc=board[index].cb;cc<=board[index].ce;++cc)
		{
			int i=cc*height+rr;
			if(board[i].mine) flag(i);
		}
}
void chord(int index)
{
	int rr,cc;
	info("Chord",index); 
	++zini;
	for(rr=board[index].rb;rr<=board[index].re;++rr)
		for(cc=board[index].cb;cc<=board[index].ce;++cc)
			reveal(cc*height+rr); /* open all neighbours */
}

int getnumber(int index)
{
	int rr,cc;
	int res=0;
	for(rr=board[index].rb;rr<=board[index].re;++rr)
		for(cc=board[index].cb;cc<=board[index].ce;++cc)
			res+=board[cc*height+rr].mine;
	return res;
}
int getadj3bv(int index)
{
	int res=0;
	int rr,cc;
	if(!board[index].number) return 1;
	for(rr=board[index].rb;rr<=board[index].re;++rr)
		for(cc=board[index].cb;cc<=board[index].ce;++cc)
		{
			int i=cc*height+rr;
			res+=(!board[i].mine && !board[i].opening); 
		}
	if(board[index].opening) ++res;
	if(board[index].opening2) ++res;
	return res;
}
void setopeningborder(int op_id,int index)
{
	if(!board[index].opening)
		board[index].opening=op_id;
	else if(board[index].opening!=op_id)
		board[index].opening2=op_id;
}
void processopening(int op_id,int index)
{
	int rr,cc;
	board[index].opening=op_id;
	for(rr=board[index].rb;rr<=board[index].re;++rr)
		for(cc=board[index].cb;cc<=board[index].ce;++cc)
		{
			int i=cc*height+rr;
			if(board[i].number)
				setopeningborder(op_id,i);
			else if(!board[i].opening)
				processopening(op_id,i);
		}		
}
void initboard()
{
	int i;
	int r,c;
	closed_cells=width*height;
	openings=0;
	/* initialize cell positions */
	for(r=0;r<height;++r)
	{
		for(c=0;c<width;++c)
		{
			int index=c*height+r;
			board[index].rb=r?r-1:r;
			board[index].re=r==height-1?r:r+1;
			board[index].cb=c?c-1:c;
			board[index].ce=c==width-1?c:c+1;
			/* YEAH! to make it all easy i'll use up to 400 mb of your ram!!! */
		}
	}
	/* initialize numbers */
	for(i=0;i<size;++i)
	{
		/* premium=-<number_of_unflagged_mines>-1(closed)-1(dbl) */
		board[i].premium=-(board[i].number=getnumber(i))-2;
	}
	/* initialize openings */
	for(i=0;i<size;++i)
		if(!board[i].number && !board[i].opening)
			processopening(++openings,i);
	bbbv=openings;
	for(i=0;i<size;++i)
	{
		if(!board[i].opening && !board[i].mine)	++bbbv;
		board[i].premium+=getadj3bv(i);
	}
	
	/* if it's random zini, make storage */
	if(rz) maxpr=(int*)malloc(sizeof(int)*size);
}

void printboard()
{
	int r,c;
	for(r=0;r<height;++r)
	{
		for(c=0;c<width;++c)
			if(board[c*height+r].mine)
				printf("%c",'*');
			else
				printf("%d",board[c*height+r].number);
		printf("\n");
	}
}

void printadj3bv()
{
	int r,c;
	for(r=0;r<height;++r)
	{
		for(c=0;c<width;++c)
			printf("%d",getadj3bv(c*height+r));
		printf("\n");
	}
}
void applyzini()
{
	int maxp=-1;
	int curi=-1;
	int i;
	if(!rz)
		for(i=0;i<size;++i)
		{
			if(board[i].premium>maxp &&
				!board[i].mine &&
				(board[i].opened || !human))
			{
				maxp=board[i].premium;
				curi=i;
			}
		}
	else
	{
		for(i=0;i<size;++i)
		{
			if(board[i].premium>maxp &&
				!board[i].mine &&
				(board[i].opened || !human))
			{
				maxp=board[i].premium;
				curi=1;
				maxpr[0]=i;
			}
			else if(maxp>=0 && board[i].premium==maxp)
			{
				maxpr[curi++]=i;
			}
		}
		if(maxp>=0) curi=maxpr[rand()%curi];
	}
	if(curi!=-1) /* there's non-negative premium */
	{
		if(infos) printf("premium %d\n",board[curi].premium);
		if(!board[curi].opened) click(curi);
		flagaround(curi);
		chord(curi);
	}
	else /* left-click */
	{
		int i;
		for(i=0;i<size;++i)
			if(!board[i].opened && !board[i].mine &&
				(!board[i].number || !board[i].opening))
			{
				curi=i;
				break;
			}
		if(rz)
		{
			maxpr[0]=curi;
			curi=1;
			for(i=maxpr[0]+1;i<size;++i)
			{
				if(!board[i].opened && !board[i].mine &&
					(!board[i].number || !board[i].opening))
					maxpr[curi++]=i;
			}
			curi=maxpr[rand()%curi];
		}
		click(curi);
	}
}

void hitopenings()
{
	int j;
	for(j=0;j<size;++j)
		if(!board[j].number && !board[j].opened)
			click(j);
}

void zinialg()
{
	zini=0;
	if(hitops) hitopenings();
	while(closed_cells>mines)
	{
		applyzini();
		if(turns)
			pause();
		else if(infos)
			printf("\n");
	}
}

void readoptions(const char* opts)
{
	int s=0;
	while(opts[s])
	{
		switch(opts[s])
		{
		case 'h':
			hitops=human=1;break;
		case 'o':
			hitops=1;break;
		case 'r':
			rz=1;break;
		case 's':
			infos=0;break;
		case 't':
			turns=1;break;
		case '-':
			break;
		case 'b':
			nobrd=1;break;
		case 'a':
			a3=1;break;
		case 'i':
			priter=0;break;
		default:
			printf("Unknown option %c\n",opts[s]);
		}
		++s;
	}
}

int main(int argc,char** argv)
{
	const char* filename=argv[argc-1];
	int i,l=strlen(filename);

	a3=nobrd=human=rz=hitops=turns=s_iter=0;
	priter=infos=1;
	n_iter=1;

	if(argc<2)
	{
		printhelp();
		printf("Press enter to exit\n");
		pause();
		return 0;
	}

	printf("This is ZiNi calculator, version %.1f\n",ZNC_VERSION);
	/* read options */
	if(argc>2) /* assuming argv[1] is options */
		readoptions(argv[1]);
	if(argc>3) /* assuming argv[2] is number of iterations */
		n_iter=atoi(argv[2]);
	if(argc>4)
		s_iter=atoi(argv[3]);
	if(n_iter<1)
	{
		printf("Invalid number of iterations. Assuming it's 1\n");
		n_iter=1;
	}

	/* open file */
	f=fopen(filename,"rb");
	if(!f)
		error("Cannot open input file");

	/* determine file format */
	for(i=l-1;i>=0 && filename[i]!='.';--i);

	/* read board from file */
	printf("%s\n",filename+i);
	if(!strcmp(filename+i,".mbf"))
		readmbf();
	else if(!strcmp(filename+i,".abf"))
		readmbf();
	else if(!strcmp(filename+i,".mvf"))
		readmvf();
	else if(!strcmp(filename+i,".avf"))
		readavf();
	else if(!strcmp(filename+i,".umf"))
		readrmv();
	else if(!strcmp(filename+i,".rmv"))
		readrmv();
	else
		error("Unknown extension");
	fclose(f);

	initboard();
	if(!nobrd) {printboard();printf("\n");}
	if(a3) {printadj3bv();printf("\n");}
	printf("3BV: %d\n",bbbv);
	printf("Openings: %d\n",openings);

	/* count ZiNi */
	if(n_iter>1)
	{
		int min=size;
		int min_it=-1;
		int i;
		for(i=1;i<=n_iter;++i)
		{
			if(s_iter) infos=i==s_iter;
			if(s_iter && infos) priter=1;

			if(priter) printf("\nIteration %d\n",i);
			zinialg();
			if(priter) printf("ZiNi value: %d\n",zini);
			if(zini<min) {min=zini;min_it=i;}
			restartboard();
			initboard();
		}
		printf("\nMinimal ZiNi value: %d (iteration %d)\n",min,min_it);
	}
	else
	{
		zinialg();
		printf("ZiNi value: %d\n",zini);
	}

	printf("Press enter to exit\n");
	pause();
	return 0;
}
